from zeroconf import IPVersion, ServiceStateChange
from zeroconf.asyncio import AsyncServiceBrowser, \
        AsyncServiceInfo, AsyncZeroconf

from typing import cast
import asyncio


state_map = {
        ServiceStateChange.Added: "added",
        ServiceStateChange.Removed: "removed",
        ServiceStateChange.Updated: "updated"}


async def mdns_state_change(o, id, zeroconf, service_type, name, state_change):
    o.write_msg(
                id=id,
                data={
                    "service-name": name,
                    "service-type": service_type,
                    "state-change": state_map[state_change]})


class MDNS():
    async def start(self, out):
        self.out = out
        self.aiozc = AsyncZeroconf(ip_version=IPVersion.V4Only)
        self.browsers = {}
        return self

    async def info(self, id, opts):
        info = AsyncServiceInfo(opts['service-type'], opts['service-name'])
        await info.async_request(self.aiozc.zeroconf, 3000)

        data = None
        if info:
            data = {**opts, **{
                    'addresses': [
                        [add, cast(int, info.port)]
                        for add in info.parsed_scoped_addresses()],
                    'weight': info.weight,
                    'priority': info.priority,
                    'server': info.server}}
            if info.properties:
                data['properties'] = {}
                for key, value in info.properties.items():
                    data['properties'][key.decode('utf-8')] =\
                        value.decode('utf-8')

        self.out.write_msg(id=id, data=data)

    async def listen(self, id, opts):
        service = opts['service-type']

        if service not in self.browsers:

            def wrapper(**kwds):
                asyncio.ensure_future(mdns_state_change(self.out, id, **kwds))

            self.browsers[service] = AsyncServiceBrowser(
                self.aiozc.zeroconf, service, handlers=[wrapper]
            )





from datetime import datetime, timedelta
from celery.schedules import crontab
import pytz

from astral.sun import sun
from astral import LocationInfo


# l = LocationInfo('name', 'region', 'timezone/name', 0.1, 1.2)

def only_next(n, o, a, b):

    if (a+o) < n:
        c = (b+o)
    else:
        c = (a+o)

    return (c - n).total_seconds() * 1000


def astral_next_events(lat, lon, tz, offset_seconds):

    city = LocationInfo(None, None, tz, lat, lon)
    tz = pytz.timezone(tz)
    now = datetime.now(tz=tz)
    offset = timedelta(seconds=offset_seconds)

    today = sun(city.observer, date=now, tzinfo=tz)
    tomorrow = sun(city.observer, date=now + timedelta(days=1), tzinfo=tz)

    return {
            'dawn':     only_next(now, offset, today['dawn'], tomorrow['dawn']),
            'sunrise':  only_next(now, offset, today['sunrise'], tomorrow['sunrise']),
            'noon':     only_next(now, offset, today['noon'], tomorrow['noon']),
            'sunset':   only_next(now, offset, today['sunset'], tomorrow['sunset']),
            'dusk':     only_next(now, offset, today['dusk'], tomorrow['dusk'])
            }


def schedule_to_millis(now, s):
    re = s.remaining_estimate(now)
    assert type(re) is timedelta
    return s.remaining_estimate(now).total_seconds() * 1000


def nowfun(tz):
    return datetime.now(tz=pytz.timezone(tz))


def ms_until_crontab(c):
    tz = c.pop("tz")
    return schedule_to_millis(nowfun(tz), crontab(**c))


def ms_until_solar(s):
    d = astral_next_events(s['lat'], s['lon'], s['tz'],
                           s['offset-seconds'] if 'offset-seconds' in s else 0)
    return d[s['event']]



import json
import websockets
import aiohttp


def make_host(opts):
    return f"http://{opts['host']}:{opts['port'] if 'port' in opts else 80}/api/{opts['api-key']}"


class Deconz:
    def __init__(self):
        self.uri = self.ws = None

    async def put(self, out, id, opts, type_name, device_id, state):

        ep = f'{make_host(opts)}/{type_name}/{device_id}/state'

        async with aiohttp.ClientSession() as session:
            async with session.put(ep, json=state) as response:
                response_data = await response.json()
                out.write_msg(id=id, data=dict(success=response.ok, data=response_data))

    async def listen(self, out, id, opts):
        config = make_host(opts)

        async with aiohttp.ClientSession() as session:
            async with session.get(config) as response:
                config_data = await response.json()
                out.write_msg(id=id, data=config_data)

                self.uri = "ws://{}:{}".format(
                        opts['host'],
                        config_data['config']['websocketport'])

        if self.uri:
            self.ws = await websockets.connect(self.uri)

            try:
                while True:
                    if not self.ws.open:
                        self.ws = await websockets.connect(self.uri)
                    else:
                        data = await self.ws.recv()
                        out.write_msg(id=id, data=json.loads(data))

            except websockets.exceptions.ConnectionClosedOK:
                pass



from asyncio_mqtt import Client


async def mqtt_listen(out, id, opts):

    client = Client(opts['host'])

    if 'username' in opts or 'password' in opts:
        client._client.username_pw_set(
                opts['username'] if 'username' in opts else None,
                opts['password'] if 'password' in opts else None)

    async with client:
        async with client.unfiltered_messages() as messages:
            await client.subscribe("#")
            async for message in messages:
                out.write_msg(
                        id=id,
                        data={
                            "topic": message.topic,
                            "payload": message.payload.decode()})



from . import mdns, esphome, deconz, mqtt, schedule
import asyncio
import traceback
import json
import sys
from boltons.iterutils import remap
from bcoding import bencode, bdecode


def eprint(e):
    sys.stderr.buffer.write(e.encode('utf-8') + '\n'.encode('utf-8'))
    sys.stderr.buffer.flush()


def make_clj_code(fn_name):
    return [{"name": '{}*'.format(fn_name)},
            {"name": fn_name, "code": """
(defn """ + fn_name + """
  ([opts cb _])
  ([opts cb]
   (babashka.pods/invoke
    "pod.xlfe.radiale"
    'pod.xlfe.radiale/""" + fn_name + """*
    [opts]
    {:handlers
     {:success (fn [event] (cb (assoc event
                                 :opts (dissoc opts :password :api-key)
                                 :fn-name :""" + fn_name + """)))
      :error (fn [{:keys [:ex-message :ex-data]}]
               (binding [*out* *err*]
                 (println "ERROR:" ex-message)))}})
   nil))
"""}]


def describe_this(fn_names):
    d = [{"name": "sleep-ms"}]
    for fs in fn_names:
        d.extend(make_clj_code(fs))
    return {"format": "json",
            "ops": {"shutdown": {}},
            "namespaces":
            [{"name": "pod.xlfe.radiale",
              "vars": d}]}


def _write_(d):
    bencode(d, sys.stdout.buffer)
    sys.stdout.buffer.flush()


def clean_data(path, key, value):
    return value is not float("nan") and \
           value is not float("inf") and \
           value is not float("-inf")


class OutgoingQ():

    async def start(self):
        self.running = True
        self.outgoing = asyncio.Queue()
        asyncio.create_task(self.out_task(), name="Output writer")
        return self

    async def out_task(self):
        while self.running:
            d = await self.outgoing.get()
            await asyncio.get_event_loop().run_in_executor(None, _write_, d)

    def write_raw(self, d):
        self.outgoing.put_nowait(d)

    def write_msg(self, id, data, status="status"):
        self.write_raw(
            dict(
                    value=json.dumps(
                        remap(data, visit=clean_data)
                        if type(data) is dict else data
                        ),
                    id=id,
                    status=[status])
                )


class RadialePod(object):

    def __init__(self):
        self.running = True
        self.services = {}
        self.options = {}
        self.esp_clients = {}

    async def run_pod(self):
        self.out = await OutgoingQ().start()

        while self.running:
            msg = await asyncio.get_event_loop().\
                    run_in_executor(None, bdecode, sys.stdin.buffer)

            op = msg["op"]

            if op == "describe":
                self.out.write_raw(
                        describe_this([
                            'listen-mdns',
                            'listen-mqtt',
                            'listen-deconz',
                            'millis-solar',
                            'millis-crontab',
                            'put-deconz',
                            'mdns-info',
                            'subscribe-esp',
                            'switch-esp',
                            'light-esp'])
                        )

            elif op == 'shutdown':
                self.running = self.out.running = False

            elif op == "invoke":
                try:
                    await self.invoke(msg)
                except Exception:
                    eprint(traceback.format_exc())
                    ex_type, ex_value, ex_traceback = sys.exc_info()
                    self.out.write_msg(
                            id=msg["id"],
                            status="error",
                            data=repr(ex_value))


    async def invoke(self, msg):
        var = msg["var"]
        id = msg["id"]
        opts = json.loads(msg["args"])[0]

        if var.endswith('sleep-ms'):
            await asyncio.sleep(int(opts)/1000.0)
            self.out.write_msg(id=id, status="done", data=opts)

        elif var.endswith('listen-mdns*'):
            if 'mdns' not in self.services:
                self.services['mdns'] = \
                        await mdns.MDNS().start(self.out)
            asyncio.create_task(self.services['mdns'].listen(id, opts))

        elif var.endswith('mdns-info*'):
            assert 'mdns' in self.services
            asyncio.create_task(self.services['mdns'].info(id, opts))

        elif var.endswith('listen-deconz*'):
            assert 'deconz' not in self.services
            self.services['deconz'] = deconz.Deconz()
            self.options['deconz'] = opts
            asyncio.create_task(
                self.services['deconz'].listen(self.out, id, opts))

        elif var.endswith('put-deconz*'):
            assert 'deconz' in self.services
            type_name = opts['type']
            device_id = opts['id']
            state = opts['state']
            opts = self.options['deconz']

            asyncio.create_task(
                self.services['deconz'].put(
                    self.out, id, opts,
                    type_name, device_id, state
                    ))

        elif var.endswith('listen-mqtt*'):
            asyncio.create_task(
                    mqtt.mqtt_listen(self.out, id, opts))

        elif var.endswith('subscribe-esp*'):
            self.esp_clients.update(
                await asyncio.create_task(
                    esphome.subscribe_esp(self.out,
                        self.services['mdns'].aiozc.zeroconf, id, opts))
            )
        elif var.endswith('switch-esp*'):
            client = self.esp_clients[opts['service-name']]
            assert client
            await esphome.switch_command(self.out, id, client, opts['key'], opts['state'])

        elif var.endswith('light-esp*'):
            client = self.esp_clients[opts['service-name']]
            assert client
            await esphome.light_command(self.out, id, client, opts['key'], opts['params'])

        elif var.endswith('millis-solar*'):
            ms = schedule.ms_until_solar(opts)
            while ms < 1000:
                await asyncio.sleep(1)
                ms = schedule.ms_until_solar(opts)

            self.out.write_msg(id=id, status="done", data=dict(ms=ms))

        elif var.endswith('millis-crontab*'):
            ms = schedule.ms_until_crontab(opts)
            while ms < 1000:
                await asyncio.sleep(1)
                ms = schedule.ms_until_solar(opts)

            self.out.write_msg(id=id, status="done", data=dict(ms=ms))






import pychromecast
import asyncio


class Chromecast:
    def __init__(self, kernel):
        pass

class GoogleHome:
    """
        Create a Google home (an host or a devicename is mandatory)
        :param devicename: string : the ip or device name of the Google Home
        :param host: the host of google home
        :param port: the port to contact google home by ip (default is 8009)
        :param ttsbuilder: function: the tts function. This is a function who have two parameter a text string and the lang. This function return an url to download mp3 file
    """
    def __init__(self, devicename=None):
        chromecasts, browser = pychromecast.get_listed_chromecasts(friendly_names=[devicename])
            
        self.cc = chromecasts[0]

    def say(self, text):
        tts = gTTS(text, lang='en', tld='com.au')
        tts.save('hello.mp3')

    def play(self, url, contenttype = 'audio/mp3'):
        self.cc.wait()
        mc = self.cc.media_controller
        mc.play_media(url, contenttype)
        mc.block_until_active()
        print("played url " + url)




import aioesphomeapi
from aioesphomeapi.reconnect_logic import ReconnectLogic
from aioesphomeapi.core import APIConnectionError


async def subscribe_esp(out, zc, id, opts):
    host, port = opts['addresses'][0]
    cli = aioesphomeapi.APIClient(host, port, None)

    # await cli.connect(login=True)
    def esp_change_callback(state):
        out.write_msg(
                id=id,
                data={
                    "service-name": opts['service-name'],
                    "state": [str(state.key), state.state]
                    })

    async def on_disconnect():
        out.write_msg(id=id, data={
                    "service-name": opts['service-name'],
                    "connected": False
                    })

    async def on_connect():
        sensors = await cli.list_entities_services()
        service_details = {}

        for s in sensors[0]:
            service_details[str(s.key)] = \
                {**s.to_dict(), **{"type": "service"}}

        for s in sensors[1]:
            service_details[str(s.key)] = \
                {**s.to_dict(), **{"type": "user-defined-service"}}

        out.write_msg(
                id=id,
                data={
                    "service-name": opts['service-name'],
                    "services": service_details
                    }
                )

        try:
            await cli.subscribe_states(esp_change_callback)
            out.write_msg(id=id, data={
                    "service-name": opts['service-name'],
                    "connected": True
                    })

        except APIConnectionError:
            cli.disconnect()

    logic = ReconnectLogic(
        client=cli,
        on_connect=on_connect,
        on_disconnect=on_disconnect,
        zeroconf_instance=zc
    )
    await logic.start()
    return {opts['service-name'].split('.')[0]: cli}


async def switch_command(out, id, client, key, state):
    await client.switch_command(key, state)
    out.write_msg(id=id, data={"success": True})


async def light_command(out, id, client, key, params):
    await client.light_command(key, **params)
    out.write_msg(id=id, data={"success": True})





from distutils.core import setup

setup(
    name='radiale',
    packages=['radiale'],
    version='0.5.1',
    license='EPL',
    description='radiale',
    url='https://github.com/xlfe/radiale',
    keywords=['home-automation'],
    install_requires=[
        'websockets',
        'aiohttp',
        'pychromecast',
        'aioesphomeapi',
        'zeroconf',
        'bcoding',
        'asyncio-mqtt',
        'boltons',
        'celery',
        'astral',
        'pytz'
    ],
    classifiers=[
     'Development Status :: 3 - Alpha',
     'Intended Audience :: Developers',
     'Topic :: Home Automation',
     'License :: OSI Approved :: Eclipse Public License 2.0 (EPL-2.0)',
     'Programming Language :: Python :: 3.9',
    ]
)

